#!/usr/bin/perl -w

use strict;
use threads;
use Getopt::Long;
use Pod::Usage;

sub install_bins;
sub code;
sub precompute;
sub call_rnafold_threads;
sub call_rnafold;
sub compute_pairwise_aligns;
sub extract_score_list;
sub compute_multiple_aligns;
sub copy;
sub SECTION;
sub system_command;
sub system_pipe;

###############################################################################################
#
# Global Variables and Command Line Options
#
###############################################################################################
my @binaries = ("rnaclustDecodeTable", "rnaclustScores2Dist.pl", "rnaclustGenReport.pl", 
		"rnaclustAlignRange.pl", "pgma", "locarna", "mlocarna", "RNAalifold", "RNAfold", "RNAz");

my $namesfile="names";      # relative to $tgtdir
my $help;
my $man;
my $multipleAligns=0;
my $tree=0;
my $start=-1;  
my $end=-1;
my $partition=0;          # Are start and end alignments given?
my $fasta_file="";
my $tgtdir="";
my $numberThreads=1;      # by default allow just one thread for locarna and RNAfold calls
my $N=0;
my $locarna_options="";
my $mlocarna_options="";
my $dponly=0;
my $dpdir="";
my $verbose=0;	 	  # Report all third party software calls?
my $printVersion=0;

GetOptions(	   
	   "verbose:i"=>\$verbose,
	   "help"=> \$help,
	   "man" => \$man,
	   "tree" => \$tree,
	   "malig" => \$multipleAligns,
	   "start:i" => \$start,
	   "end:i" => \$end,
	   "fasta:s" => \$fasta_file,
	   "dir:s" => \$tgtdir,
 	   "locarna-opts=s" => \$locarna_options,
	   "mlocarna-opts=s" => \$mlocarna_options,
           "cpu:i" => \$numberThreads,
           "dpdir:s" => \$dpdir,
           "dponly" => \$dponly,
	   "version" => \$printVersion
	   ) || pod2usage(2);

if($printVersion){ print STDOUT "$0 version 1.0beta\n";  exit(1); }
if($help){ pod2usage(1); }
if($man){ pod2usage(-exitstatus => 0, -verbose => 2); }
if(($fasta_file eq "" && !($tree || $multipleAligns)) || $tgtdir eq ""){ pod2usage(1);  }
if(($start>-1 && $end==-1) || ($start==-1 && $end>-1)){ pod2usage(1); exit(-1); }
if($start>-1 && $end>-1){ $partition=1; }
if($multipleAligns && $partition){ pod2usage(1); exit(-1); }
if($multipleAligns && $dponly){ pod2usage(1); exit(-1); }

if( -e $tgtdir && !$tree && !$partition && !$multipleAligns && ($dpdir eq "")) {
    print STDERR "Target directory $tgtdir already exists.\n";
    print STDERR "Please choose a different target directory (or delete it) and run again.\n";
    exit(-1);
}

###############################################################################################
#
# Main
#
###############################################################################################
if(!$tree){
	
    SECTION("Local copies of binaries");
    install_bins();
    
    if($dpdir eq ""){
	SECTION("Compute the base pair probability matrices");
	$N=precompute($fasta_file);
	if($dponly==1){ 
		SECTION("Cheers.");
		exit(1); 
	}
	$dpdir = $tgtdir."/pp";
    }
    else{
	SECTION("Use base pair probability matrices from directory $dpdir");
	$N = `grep -c '>' $fasta_file`; chomp $N;  # $fasta_file only for number of sequences needed
    }

    if($start!=-1 && $end!=-1){ SECTION("Compute the pairwise alignments $start - $end");  }
    else { SECTION("Compute the pairwise alignments"); }
    compute_pairwise_aligns();
    if($partition){ 
	SECTION("Cheers.");
	exit(1); 
    }

}
	
## Test required input files
if(!-e "$tgtdir/names"){
	# Precomputation step without generating pairwise alignments, but generating names and seqs.fasta
	precompute($fasta_file, 1)
}
if(!-e "$tgtdir/paligs/" || !-e "$tgtdir/bin/"){
	die "ERROR ($0): Target directory '$tgtdir' must contain file 'names' as well as directories 'paligs' and 'bin'.\n";
}

SECTION("Extract the score list $tgtdir");
chdir "$tgtdir";
extract_score_list();

SECTION("Generate the cluster-tree");
system_command("bin/pgma names dist-list > tree");
if($verbose){ system("cat tree"); }

chdir "../";

if($multipleAligns){

	## Test required input files
	if(!-e "$tgtdir/seqs.fasta" || !-e "$tgtdir/tree" || !-e "$tgtdir/names" || !-e "$dpdir" || !-e "$tgtdir/bin" || !-e "$tgtdir/paligs"){ 
		die "ERROR ($0): Target directory '$tgtdir' must contain files 'seqs.fasta', 'tree' and 'names', as well as directories 'paligs/', 'pp/' and 'bin/'.\n"; 
	}
	
	SECTION("Compute all multiple alignments");
	chdir "$tgtdir";
	my $maligdir=compute_multiple_aligns();

	SECTION("Generate report");
	system "bin/rnaclustGenReport.pl $maligdir > report";
	system "cp report report.copied";
	system "(echo 'ID SIZE SCORE MPI SCI LOOPS MFE LEN'; cat report.copied ) > report";
	system "rm -f report.copied";
	system "cat report";
}
chdir "..";

SECTION("Cheers.");

###############################################################################################
#
# Subs
#
###############################################################################################

sub install_bins
{
    # If bin directory exists already --> do not copy binaries a second time!
    if(-e "$tgtdir/bin"){ return; }
	
    mkdir $tgtdir;
    mkdir "$tgtdir/bin";

    # Get the full path name of the target directory
    my $tgtdir_path = `dirname $tgtdir`; chomp($tgtdir_path);
    $tgtdir_path = `cd $tgtdir_path; pwd`; chomp($tgtdir_path);    
    $tgtdir = `basename $tgtdir`; chomp($tgtdir);
    $tgtdir = "$tgtdir_path/$tgtdir"; 

    for my $f (@binaries) {
	my $t = `which $f`; chomp($t);
	if($t=~/^\s*$/){
		die "ERROR ($0): Could not find executable of '$f'!\n";
	}
        copy("$t", "$tgtdir/bin");
    }
    return;
}

sub code
{
    my ($n,$m)=@_;
    my $idx;

    $idx=0;
    for (my $i=1; $i<$n-1; $i++) {$idx+=$i;}
    $idx+=$m;

    return $idx;
}

sub precompute
{
    my ($infile, $namesOnly)=@_;   # if $namesOnly is set to 1 no pairwise alignments are created, just the names and seqs.fasta files are created
                                   # required in case dpdir exists already and user did not generate a names file himself

    my $ppdir = "$tgtdir/pp";

    local *NAMES;
    local *IN;
    
    # check if names are unique in first 13 characters
    my $addID=0;
    my @duplicates = readpipe("grep \">\" $infile | perl -ne \'print substr(\$_,1,13), \"\n\"\' | uniq -d");
    if($#duplicates>=0){
	print STDERR "WARNING ($0): Sequence names are not unique in first 13 characters, but this is required to meet\n";
	print STDERR "standard of RNAfold. A unique identifier in format 'id<number>' is added as prefix to each sequence name.\n";
	$addID=1;
    }
    
    if(!$namesOnly){
	    mkdir $ppdir || die "ERROR ($0): Cannot make dir $ppdir\n";
    }
    open(IN,$infile) || die "ERROR ($0): Cannot open $infile for reading.\n";
    
    open(NAMES,">$tgtdir/$namesfile") || die "ERROR ($0): Cannot open $tgtdir/$namesfile for writing.\n";
    my $line=<IN>;

    ## Collect RNAfold calls and pass them to different threads 
    ## Create copy of infile: seqs.fasta - each sequence on one line
    open(OUT, ">$tgtdir/seqs.fasta") || die "ERROR ($0): Cannot create file '$tgtdir/seqs.fasta'\n";
    my @aSeq=();
    my @aName=();
    my @aID=();
    my $cntID=1;
    my $cntSeq=0;
    while($line) {
	
	## get name
	($line =~ /^>\s*(\S+)/) || die "ERROR ($0): Expected name.\n";
	my $name = $1;
	# Replace non-word characters
	$name =~ s/\W/_/g;
	if($addID==1){ $name = "id$cntID"."_".$name; }
	print NAMES "$name\n";
	
	## read sequence
	my $seq="";
	while (defined($line=<IN>) && ($line !~ /^>/)) {
	    if($line =~ /^\s*$/){ next; } # ignore empty lines	
	    $line =~ s/\s//g;
	    $seq.=$line;
	};

	$cntSeq++;
	print OUT ">$name\n$seq\n";
	push(@aSeq, $seq);
	push(@aName, $name);
	push(@aID, $cntID++);

	## Call numberThread threads
	if(!$namesOnly){
	        if($cntSeq==$numberThreads){	
		    call_rnafold_threads(\@aSeq, \@aName, \@aID, $ppdir, $numberThreads);
		    @aSeq=();   # free 
		    @aName=();
		    @aID=();
		    $cntSeq=0;  # start collecting again
        	}
	}
    }
    close(OUT);

    ## Call the remaining threads
    if(!$namesOnly){
	    call_rnafold_threads(\@aSeq, \@aName, \@aID, $ppdir, $#aName+1);
    }
        
    close NAMES;
    close IN;
    return $cntID-1;  # subtract 1, otherwise code $N $N-1 calculates wrong number of pairwise alignments
}

sub call_rnafold_threads
{
    my ($ref_aSeq, $ref_aName, $ref_aID, $ppdir, $nthreads) = @_;

    my @threads;

    if($nthreads == 1){  # one cpu only --> discard use of Perl threads
	call_rnafold($ref_aName->[0], $ref_aSeq->[0], $ref_aID->[0], $ppdir);
    }
    else{
	## Call threads here
	for(my $i=0; $i<$nthreads; $i++){
	    $threads[$i]=threads->create(\&call_rnafold, $ref_aName->[$i], $ref_aSeq->[$i], $ref_aID->[$i], $ppdir);
	}
	## Wait until they are ready
	for(my $i=0; $i<$nthreads; $i++){
	    my $r = $threads[$i]->join();
	}
    }
    return;
}

sub call_rnafold
{
    my ($name, $seq, $id, $ppdir) = @_;

    # RNAfold ignores everything after the 13th character in name
    $name = substr($name,0,12);
    $name =~ s/\//_/g; 
    $name =~ s/\./_/g;
    my $pp_outfile = "$ppdir/$id";
    my $dpplot = $name."_dp.ps";
    my $dpplot2 = $name."_dp2.ps";
    my $rnaplot = $name."_ss.ps";
    if($locarna_options=~/--stacking/){
	# If stacking is regarded locarna expects RNAfold -p2 dotplots
	system_pipe("echo -e \">$name\n$seq\" | RNAfold -p2 -d2");
	copy "$dpplot2", "$pp_outfile";
	unlink "$rnaplot"; unlink "$dpplot"; unlink "$dpplot2";
    }
    else{
	system_pipe("echo -e \">$name\n$seq\" | RNAfold -p -d2");	
	copy "$dpplot", "$pp_outfile";
	unlink "$rnaplot"; unlink "$dpplot";
    }
    return;
}

sub compute_pairwise_aligns
{
#   my $alignRange = new RNAclust::rnaclustAlignRange($locarna_options, "$tgtdir/bin/locarna", $tgtdir, $numberThreads);
    if(!$partition){
	$N=code $N, $N-1;
	print "  be patient, we compute $N alignments ...\n";
	#$alignRange->align_range("1-$N");
	system_command("$tgtdir/bin/rnaclustAlignRange.pl --dpdir $dpdir --range 1-$N --locarna-path $tgtdir/bin/locarna --tgtdir $tgtdir --cpu $numberThreads --verbose $verbose --locarna-opts \"$locarna_options\"");
	unlink "alirna.ps";
    }
    else{
	#if($end==-1 && $step!=-1){ $end = $start+$step-1; }
	my $maxN=code $N, $N-1;
	if($end>$maxN){ die "ERROR ($0): '--end $end' exceeds maximal number of alignments, wich is $maxN.\n"; }
	$N=$end-$start+1;
	print "  be patient, we compute $N alignments ...\n";
#	$alignRange->align_range("$start-$end");
	system_command("$tgtdir/bin/rnaclustAlignRange.pl --dpdir $dpdir --range $start-$end --locarna-path $tgtdir/bin/locarna --tgtdir $tgtdir --cpu $numberThreads --verbose $verbose --locarna-opts \"$locarna_options\"");
	unlink "alirna.ps";
	exit;
    }
    return;
}

sub extract_score_list
{
    print "bunzip2 ...\n";
    open(IN,"bunzip2 -c paligs/*.bz2 |") || die "Cannot read from $!";
    open(OUT,">score-list") || die "Cannot write $!";
    
    # @ID and Score etc. are now found on same line
    print "generate score list ...\n";
    my $foundScores=0;
    while(my $line=<IN>) {
	if($line =~ /^\@(\d+)/){
	    print OUT "$1 ";
	}
	#if($line =~ /Score:\s+([\-\d\.]+)\s+SCI:\s+([\d\.\-]+)\s+ID:\s+([\d\.\-]+)\s+([\d\.\-]+)\s*$/) {
	if($line =~ /Score:\s+([\-\d\.]+)\s*$/) {  # sci and identity are actually not needed --> saves computation time
	    # $1: Score $2: SCI $3: sequence identity $4: sequence identity of non-structural positions
	    print OUT "$1\n";
	    $foundScores++;
	}
    }
    close IN;
    close OUT;
    if($foundScores==0){
	die "ERROR ($0): Failed to extract Score!\n";
    }
    print "$foundScores scores have been extracted\n";
   
    print "decode table ...\n";
    system_command("bin/rnaclustDecodeTable < score-list > score-list.dec");

    rename "score-list.dec", "score-list";

    print "scores2dist ...\n";
    system_command("bin/rnaclustScores2Dist.pl < score-list > dist-list");
    
    return;
}

sub compute_multiple_aligns
{
    open(NAMES,"names") || die "Cannot read 'names' file!";
    my $maligdir="maligs";
    mkdir "$maligdir.mloc";
    chdir "$maligdir.mloc";
    my $i=0;
    while (my $name=<NAMES>) {
	$i++;
	my $short_name = sprintf "%16s",$name;
	$short_name =~ s/[^A-Za-z0-9]/_/g;
	system("ln -s ../pp/$i $short_name");
    }
    close NAMES;
    chdir "..";
    
    #my $incl = `pwd`; chomp($incl);
    system_command("$tgtdir/bin/mlocarna --treefile tree --tgtdir $maligdir.mloc --skip-pp $mlocarna_options seqs.fasta > mlocarna.out");
    
    opendir(DIR,"$maligdir.mloc");
    my @clusterfiles = readdir(DIR);
    closedir DIR;
    
    mkdir "$maligdir";
    for my $f (@clusterfiles) {
	if ($f =~ /^intermediate(\d+).aln$/) {
	    my $idx=$1;
	    rename "$maligdir.mloc/$f","$maligdir/$idx.aln";
	}
    }

    # clean up
    system "rm -rf $maligdir.mloc";

    return $maligdir;
}

###############################################################################################
#
# Helpers
#
###############################################################################################

sub copy{ system("cp @_"); }   
 
sub SECTION
{
    my ($s)=@_;
    print "============================================================\n";
    print "$s\n";
    print "------------------------------------------------------------\n";
}

sub system_command
{
    my($sysCommand) = @_;

    if($verbose){ print "CALL $sysCommand\n"; }
    system("$sysCommand")==0
        or die "ERROR ($0): Could not call system command '$sysCommand'!\n\n";
    return 1;
}

sub system_pipe
{
    my($sysCommand) = @_;

    if($verbose){ print "CALL $sysCommand (readpipe)\n"; }
    my @lines = readpipe("$sysCommand");
    return 1;
}


###############################################################################################
#
# Docu
#
###############################################################################################

__END__

=head1 NAME

RNAclust.pl

=head1 SYNOPSIS

RNAclust.pl [options]

Options:

    --fasta	<file name>			FASTA file
						(required)

    --dir	<dir name>			Target directory TGTDIR
						(required)

    --cpu       <integer>      		   	Number of CPUs available on your machine. Calls to locarna and 
                                		RNAfold will be distributed equally on those CPUs. (default: 1)
                                                (optional)

    --dponly                                    Stop after Dotplots are created. This option is useful if 
                                                calculation of pairwise alignments will be distributed among 
                                                different machines. 
                                                (optional)

    --dpdir     <dir>                           Directory of dotplots. Use this option in case you distribute the 
                                                calculation of pairwise alignments among different machines using
                                                --start and --end options.
                                                (optional)

    --start     <integer>        		Start with pairwise alignment i.
						(optional)

    --end       <integer>	        	Stop with pairwise alignment j.
						(optional, except --start is specified)

    --tree					Use this option after all pairwise alignments have been computed 
						on different machines (options --start, --end) in order to generate 
						the hierarchical cluster-tree. The pairwise alignments must be moved 
						to one 'paligs' directory beforehand.
						(optional)

    --malig					Final cluster-tree is traversed and multiple alignment for each internal 
						node is computed progressively, default: no 
						(optional)

    --locarna-opts <"locarna options">		options passed directly to locarna (must be given as one string)
						(optional)

    --mlocarna-opts <"mlocarna options">	options passed directly to mlocarna (must be given as one string). mlocarna
                                                is used if multiple alignments for the internal nodes will be calculated (see --malig).
						(optional)

    --help	 		        	brief help message
						(optional)

    --man       	    	        	full documentation
						(optional)

    --version					print version
	
=head1 DESCRIPTION

    RNAclust.pl is a perl script summarizing all the single steps required for structural clustering of RNA structure motifs. It requires as input
    a multiple FASTA file (the sequence names should be identical in the first 13 characters due to requirements of the Vienna RNA package).
    In the first step it calculates for each input sequence the base pair probability matrix (given as dotplot) of its secondary structure distribution 
    using RNAfold from the Vienna RNA package. Secondly, for each pair of dotplots a sequence-structure alignment is calculated using LocARNA. Lastly, 
    a hierarchical cluster-tree (in NEWICK format) is derived by WPGMA clustering of the pairwise alignment distances. 

    Calculating all pairwise sequence-structure alignments is the bottleneck of this pipeline, although comparable fast in case LocARNA is used.
    Hence, RNAclust.pl provides the possibility to distribute this step between different CPUs on your computer (see --cpu option). Furthermore, 
    by using --start and --end options the caluculation of pairwise alignments can be distributed among different machines. 

    As the post-processing of a large tree is problematic, you may use RNAsoup in order to derive those subtrees which are likely 
    to define a distinct structural motif. 

    Invocation examples
    -------------------

    (1) RNAclust.pl --fasta FASTA --dir TARGETDIR --malig

      Runs the complete clustering pipeline, i.e. calling RNAfold -p to create a dotplot for each input sequence. Dotplots are written to 
      TARGETDIR/pp. All pairwise alignments are calculated on one CPU. BZIP2 archives containing all those alignments are found in TARGETDIR/paligs.
      The cluster-tree is derived and saved in TARGETDIR/tree. Lastly, the cluster-tree is traversed and for each internal node a multiple alignment
      is calculated progressively. Multiple alignments are saved in TARGETDIR/maligs.

    (2) RNAclust.pl --fasta FASTA --dir TARGETDIR --cpu 4

      Runs the complete clustering pipeline on the same machine but distributing the calculation of base pair probability matrices (i.e. dotplots) and the 
      pairwise alignments between 4 CPUs of your machine. Does not calculate the multiple alignments for each internal node of the final cluster-tree.

    (3) RNAclust.pl --fasta FASTA --dir TARGETDIR --cpu 4 --locarna-opts "--struct-local 1 --sequ-local 1"

      As above, but two parameters are passed to LocARNA enabling full local alignments.

    (4.1) RNAclust.pl --fasta FASTA --dir TARGETDIR --dponly

      Creates only the base pair probability matrices. They are saved in TARGETDIR/pp. 

    (4.2) RNAclust.pl --fasta FASTA --dir TARGETDIR --dpdir DIR/pp --start 1000 --end 1999

      Calculates only pairwise alignments 1000 to 1999 based on the dotplots provided in DIR/pp. Stops after those pairwise alignments are 
      calculated. 

    (4.3) RNAclust.pl --dir TARGETDIR --tree

      Calculates the final cluster-tree after pairwise alignments have been computed on different machines. TARGETDIR/paligs must contain the BZIP2 files 
      of all pairwise alignments (collect them from the different machines). Furthermore TARGETDIR/names must exist.

    (4.1), (4.2) and (4.3) have to be called consecutively in order to distribute the calculation of pairwise alignments among different machines.

    For a more detailed information please read the manual.

=head1 AUTHOR

    Kristin Reiche (kristin@bioinf.uni-leipzig.de) and Sebastian Will (will@informatik.uni-freiburg.de)

=cut
